#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""PySimpleSOAP"""


__author__ = "Mariano Reingart"
__author_email__ = "reingart@gmail.com"
__copyright__ = "Copyright (C) 2013 Mariano Reingart"
__license__ = "LGPL 3.0"
__version__ = "1.16.2"

TIMEOUT = 60


from . import client, server, simplexml, transport
